#| events
dungeon_world_events:
    type: world
    debug: false
    events:
        # these events ensure the template world is loaded
        on server prestart:
        - createworld world_template generator:denizen:void
        - foreach <server.flag[active_dungeon_list]>:
            - createworld <[value]> generator:denizen:void
            - adjust <world[<[value]>]> destroy
        - flag server active_dungeon_list:!
        on script reload:
        - if !<world[world_template].exists>:
            - createworld world_template generator:denizen:void

        # remove the dungeon world after the last player leaves
        after player changes world from dungeon_*:
        - flag <player> dungeon:!
        - if <player.has_flag[dungeon_old_gamemode]>:
            - adjust <player> gamemode:<player.flag[dungeon_old_gamemode]>
            - flag <player> dungeon_old_gamemode:!
        - wait 1t
        - if <context.origin_world.players.is_empty.if_null[true]>:
            - waituntil rate:5s max:5m <context.origin_world.has_flag[dungeon_spawned]>
            - adjust <context.origin_world> destroy
            - flag server active_dungeon_list:<-:<context.origin_world.name>

        # reset the gamemode in case of server crashes
        after player joins flagged:dungeon_old_gamemode:
        - if <player.has_flag[dungeon_old_gamemode]>:
            - adjust <player> gamemode:<player.flag[dungeon_old_gamemode]>
            - flag <player> dungeon_old_gamemode:!
        # ensure a player is put outside of the dungeon if they quit inside one
        on player quits:
        - define world <player.location.world>
        - if <[world].starts_with[dungeon_]>:
            - teleport <player> <player.flag[dungeon_old_location]>
            - flag <player> dungeon_old_location:!
            - if <player.has_flag[dungeon_old_gamemode]>:
                - adjust <player> gamemode:<player.flag[dungeon_old_gamemode]>
                - flag <player> dungeon_old_gamemode:!
            - wait 1t
            - if <[world].players.is_empty.if_null[true]>:
                - adjust <[world]> destroy
                - flag server active_dungeon_list:<-:<[world].name>


dungeon_creation_events:
    type: world
    debug: false
    events:
        on player right clicks block with:dungeon_item:
        - if !<player.has_permission[dungeon.place]>:
            - narrate "<dark_red>you don't have permission to create dungeons"
            - determine cancelled
        # if the player clicked an actual block
        - if <context.location.exists>:
            # check if there is already a dungeon
            - if <context.location.has_flag[dungeon]>:
                - narrate "there is already a dungeon here, break the block to remove it"
                - determine cancelled
            # store the dungeon settings contained in the flag on the item
            - flag <context.location> dungeon:<context.item.flag[flag_map]>
            # spawn a marker that displays the dungeon name
            - spawn dungeon_marker <context.location.center> save:marker
            - adjust <entry[marker].spawned_entity> custom_name:<context.item.flag[flag_map].get[dungeon_name]>
            - flag <context.location> marker:<entry[marker].spawned_entity>
            - narrate "created dungeon at <context.location.simple>"
        - determine cancelled

        on player breaks block location_flagged:dungeon:
        # let players with permission remove dungeons
        - if !<player.has_permission[dungeon.remove]>:
            - narrate "<dark_red>you don't have permission to remove dungeons"
            - determine cancelled
        # double check in case of accidental breaking
        - else if !<player.has_flag[break_dungeon_block]>:
            - narrate "are you sure you want to delete this dungeon?"
            - narrate "break again to remove the dungeon"
            - flag <player> break_dungeon_block expire:10s
            - determine cancelled
        - else:
            # remove all settings and the name armour stand
            - flag <context.location> dungeon:!
            - flag <player> break_dungeon_block:!
            - remove <context.location.flag[marker]>
            - flag <context.location> marker:!
            - narrate "dungeon removed"
            - determine cancelled

dungeon_spawning_events:
    type: world
    debug: false
    events:
        on player right clicks block location_flagged:dungeon:
        - determine passively cancelled
        # ratelimit so the player cant start another dungeon while one is generating
        - ratelimit <player> 10s
        - if <context.location.world.name.starts_with[dungeon_]>:
            - narrate "dungeons within dungeons are untested, use at your own risk"
            - determine cancelled
        - narrate "Entering the dungeon"
        # get the dungeon settings from the location and define some often used parameters
        - define dungeon_settings <context.location.flag[dungeon]>
        - define path_count <[dungeon_settings].get[path_count]>
        - define connect_chance <[dungeon_settings].get[connect_chance]>
        - define dungeon_name <[dungeon_settings].get[dungeon_name].replace[<&sp>].with[_]>
        - define dungeon_type <[dungeon_settings].get[dungeon_type]>

        # generate a unique uuid to guarentee that worldnames won't be the same
        - define uuid <util.random.uuid>
        - while <server.flag[active_dungeon_list].contains[dungeon_<[dungeon_name]>_<[uuid]>].if_null[false]>:
            - define uuid <util.random.uuid>
        # create the dungeon and track existing dungeons in a server flag
        - define dungeon_id dungeon_<[dungeon_name]>_<[uuid]>
        - flag server active_dungeon_list:->:<[dungeon_id]>
        - ~createworld <[dungeon_id]> copy_from:world_template generator:denizen:void
        # chunkload the area of the starting room
        # TODO change to dynamic size instead of static values to support dynamic room sizes
        - define chunk_x 0
        - repeat 2:
            - define chunk_z 0
            - repeat 2:
                - chunkload <chunk[<[chunk_x]>,<[chunk_z]>,<[dungeon_id]>]> duration:1m
                - define chunk_z:++
            - define chunk_x:++
        # set the reference point for the dungeon, all locations are relative to this point
        - define dungeon_origin <location[0,0,0,<[dungeon_id]>]>
        # TODO change to relative location to origin to support dynamic spawn location
        - define dungeon_spawn <[dungeon_origin].add[10,102,10]>
        # load the starting room schematic
        - if !<schematic[<[dungeon_type]>_starting_room].exists>:
            # check if there are rooms in case the preset was changes since the dungeon creation
            - if  <server.list_files[schematics\dungeons\<[dungeon_type]>\start].filter_tag[<[filter_value].ends_with[.schem]>].size> > 0:
                # pick a random starting room .schem file
                - define schematic_name <server.list_files[schematics\dungeons\<[dungeon_type]>\start].filter_tag[<[filter_value].ends_with[.schem]>].random.before[.schem]>
                - ~schematic load name:<[dungeon_type]>_starting_room filename:dungeons\<[dungeon_type]>\start\<[schematic_name]>
            - else:
                # remove the world and print an error message if there is no starting room schematic
                - debug error "<dark_red>ERROR: dungeon preset <[dungeon_type]> has no starting room schematic"
                - adjust <[dungeon_id]> destroy
                - flag server active_dungeon_list:<-:<[dungeon_id]>
                - determine cancelled
        # generate the starting room
        - ~schematic paste name:<[dungeon_type]>_starting_room <[dungeon_origin].with_y[100]>
        - wait 1t
        # teleport the player to the starting room and save the location they should return to
        - flag <player> dungeon_old_location:<player.location>
        - teleport <player> <[dungeon_spawn].add[0.5,1,0.5]>
        # set their gamemode to adventure if they are not in creative
        - if <player.gamemode> != creative:
            - flag <player> dungeon_old_gamemode:<player.gamemode>
            - adjust <player> gamemode:adventure
        # create an exit at the first solid block underneath the spawn location
        - define exit <[dungeon_spawn].add[0.5,1,0.5]>
        - while <[exit].material.name.ends_with[air]>:
            - define exit <[exit].below>
        # spawn an armour stand for the exit text and flag the location
        - spawn dungeon_marker <[exit].above> save:marker
        - adjust <entry[marker].spawned_entity> custom_name:dungeon<&sp>exit
        - flag <[exit]> dungeon_exit
        # update the light around the player
        - light <player.location> 1 d:1t
        - inject task_generate

        on player right clicks block location_flagged:dungeon_exit in:dungeon_*:
        # teleport the player out prioritising the place they entered from, otherwise using their bed or spawn
        - if <player.has_flag[dungeon_old_location]>:
            - teleport <player> <player.flag[dungeon_old_location]>
            - flag <player> dungeon_old_location:!
        - else:
            - if <player.bed_spawn.exists>:
                - teleport <player> <player.bed_spawn>
            - else:
                - teleport <player> <player.spawn_location>
dungeon_menu_events:
    type: world
    debug: false
    events:
        # events for configuring the dungeon in the /dungeon menu
        # uses sign prompts to get the users input


        # amount of paths from the starting room
        on player clicks path_count in dungeon_menu:
        # prompt the user for input
        - define message "number of paths?"
        - ~run sign_prompt def:<list.include_single[<element[].repeat_as_list[3].insert[<[message]>].at[2]>]> save:output
        - define user_input <entry[output].created_queue.determination.get[1]>
        # check if the request timed out or errored
        - if <[user_input].size.if_null[0]> > 1:
            - define user_input <[user_input].get[1]>
            # check if the value is valid, max 4 sides on the starting room
            - if <[user_input]> >= 0 && <[user_input]> <= 4:
                # save the new value for the setting
                - flag <player> flag_map.path_count:<[user_input]>
                # update the menu with the new values
                - inject path_menu_task
            - else:
                - narrate "invalid, value must be between 0 and 4"

        # add branches to an existing path
        on player clicks dungeon_path in dungeon_menu:
        # ensure the path has a length
        - if <player.flag[flag_map.path_<context.slot.sub[2]>.path_length]> <= 0:
            - narrate "set the path length first"
            - stop
        # save what path they are currently adding
        - flag <player> path_editing:<context.slot.sub[2]>
        # open the branch menu
        - inject branch_menu_task


        # set the path length
        on player clicks path_length in dungeon_menu:
        # prompt the user for input
        - define message "path length?"
        - ~run sign_prompt def:<list.include_single[<element[].repeat_as_list[3].insert[<[message]>].at[2]>]> save:output
        - define user_input <entry[output].created_queue.determination.get[1]>
        # check if the request timed out or errored
        - if <[user_input].size.if_null[0]> > 1:
            - define user_input <[user_input].get[1]>
            # check if the value is valid, any number
            - if <[user_input].is_integer>:
                # save the new value for the setting
                - flag <player> flag_map.path_<context.slot.sub[11]>.path_length:<[user_input]>
                # update the menu
                - inventory adjust d:<player.open_inventory> slot:<context.slot> lore:<[user_input]>
            - else:
                - narrate "invalid, must be a number"

        # set the chance for adjacent unconnected rooms to connect
        on player clicks connect_chance in dungeon_menu:
        # prompt the user for input
        - define message "connect chance?"
        - ~run sign_prompt def:<list.include_single[<element[].repeat_as_list[3].insert[<[message]>].at[2]>]> save:output
        - define user_input <entry[output].created_queue.determination.get[1]>
        # check if the request timed out or errored
        - if <[user_input].size.if_null[0]> > 1:
            - define user_input <[user_input].get[1]>
            # check if the value is valid, percentage between 0 and 100
            - if <[user_input]> >= 0 && <[user_input]> <= 100:
                # save the new value for the setting
                - flag <player> flag_map.connect_chance:<[user_input]>
                # update the menu
                - inventory adjust d:<player.open_inventory> slot:<context.slot> lore:<[user_input]>
            - else:
                - narrate "invalid, value must be between 0 and 100"

        # set the dungeon display name
        on player clicks dungeon_name in dungeon_menu:
        # prompt the user for input
        - define message "dungeon name?"
        - ~run sign_prompt def:<list.include_single[<element[].repeat_as_list[3].insert[<[message]>].at[2]>]> save:output
        - define user_input <entry[output].created_queue.determination.get[1]>
        # check if the request timed out or errored
        - if <[user_input].size.if_null[0]> > 1:
            - define user_input <[user_input].get[1]>
            # check if the name is already in use
            - if <server.flag[dungeon_names].contains[<[user_input]>].if_null[false]>:
                - narrate "invalid, name is already in use"
            - else:
                # save the new value for the setting
                - flag <player> flag_map.dungeon_name:<[user_input]>
                # update the menu
                - inventory adjust d:<player.open_inventory> slot:<context.slot> lore:<[user_input]>


        # set the dungeon preset type
        on player clicks dungeon_type in dungeon_menu:
        # prompt the user for input
        - define message "dungeon type?"
        - ~run sign_prompt def:<list.include_single[<element[].repeat_as_list[3].insert[<[message]>].at[2]>]> save:output
        - define user_input <entry[output].created_queue.determination.get[1]>
        # check if the request timed out or errored
        - if <[user_input].size.if_null[0]> > 1:
            - define user_input <[user_input].get[1]>
            # check if the type is valid
            - if <server.list_files[schematics\dungeons].contains[<[user_input]>]>:
                # verify that the type has at least 1 room for each type and rotation
                - if <proc[valid_type_proc].context[<[user_input]>]> == success:
                    # save the new value for the setting
                    - flag <player> flag_map.dungeon_type:<[user_input]>
                    # update the menu
                    - inventory adjust d:<player.open_inventory> slot:<context.slot> lore:<[user_input]>
            - else:
                - narrate "invalid, dungeon type does not exist"

        # set the amount of branches
        on player clicks branch_amount in path_menu:
        # prompt the user for input
        - define message "number of branches?"
        - ~run sign_prompt def:<list.include_single[<element[].repeat_as_list[3].insert[<[message]>].at[2]>]> save:output
        - define user_input <entry[output].created_queue.determination.get[1]>
        # check if the request timed out or errored
        - if <[user_input].size.if_null[0]> > 1:
            - define user_input <[user_input].get[1]>
            # check if the value is valid, max 15 slots in the menu
            - if <[user_input]> <= 15 && <[user_input]> >= 0:
                # save the new value for the setting
                - flag <player> flag_map.path_<player.flag[path_editing]>.branch_amount:<[user_input]>
                # update the menu
                - inject branch_menu_task
            - else:
                - narrate "invalid, value must be between 0 and 15"

        # set the room at which the branch will be tried
        on player clicks branch_location in path_menu:
        # prompt the user for input
        - define message "room number of branch?"
        - ~run sign_prompt def:<list.include_single[<element[].repeat_as_list[3].insert[<[message]>].at[2]>]> save:output
        - define user_input <entry[output].created_queue.determination.get[1]>
        # check if the request timed out or errored
        - if <[user_input].size.if_null[0]> > 1:
            - define user_input <[user_input].get[1]>
            # check if the value is valid, must be a number that is on the path
            - define path_length <player.flag[flag_map.path_<player.flag[path_editing]>.path_length]>
            - if <[user_input]> <= <[path_length]>:
                # save the new value for the setting on the path the player is currently editing
                - define branch <context.slot.div_int[18].mul[5].add[<context.slot.mod[18].sub[2]>]>
                - flag <player> flag_map.path_<player.flag[path_editing]>.branch_<[branch]>.location:<[user_input]>
                # update the menu
                - inventory adjust d:<player.open_inventory> slot:<context.slot> lore:<[user_input]>
            - else:
                - narrate "room number is higher than path length, room does not exist"

        # set the branch length
        on player clicks branch_length in path_menu:
        # prompt the user for input
        - define message "branch length?"
        - ~run sign_prompt def:<list.include_single[<element[].repeat_as_list[3].insert[<[message]>].at[2]>]> save:output
        - define user_input <entry[output].created_queue.determination.get[1]>
        # check if the request timed out or errored
        - if <[user_input].size.if_null[0]> > 1:
            - define user_input <[user_input].get[1]>
            # check if the value is valid, any number
            - if <[user_input].is_integer>:
                # save the new value for the setting on the path the player is currently editing
                - define branch <context.slot.div_int[18].mul[5].add[<context.slot.mod[18].sub[11]>]>
                - flag <player> flag_map.path_<player.flag[path_editing]>.branch_<[branch]>.length:<[user_input]>
                # update the menu
                - inventory adjust d:<player.open_inventory> slot:<context.slot> lore:<[user_input]>
            - else:
                - narrate "invalid, must be a number"

        # open the dungeon menu
        on player clicks dungeon_confirm in path_menu:
        - inject path_menu_task

        # verify the settings and give the dungeon item if they are valid
        on player clicks dungeon_confirm in dungeon_menu:
        - define flag_map <player.flag[flag_map]>
        # check the amount of paths
        - if <[flag_map].get[path_count]> < 1:
            - narrate "create at least 1 path first"
            - stop
        # check each path for length and branches
        - repeat <[flag_map].get[path_count]>:
            - define path_variables <[flag_map].get[path_<[value]>]>
            - define path <[value]>
            - if <[path_variables].get[path_length]> > 0:
                - if <[path_variables].get[branch_amount]> > 0:
                    - repeat <[path_variables].get[branch_amount]>:
                        - define branch_variables <[path_variables].get[branch_<[value]>]>
                        # check if branch location and length are set
                        - if <[branch_variables].get[location].if_null[0]> == 0:
                            - narrate "invalid branch location in path <[path]> for branch <[value]>"
                            - stop
                        - else if <[branch_variables].get[length].if_null[0]> == 0:
                            - narrate "invalid branch length in path <[path]> for branch <[value]>"
                            - stop
            - else:
                - narrate "unset path length for path <[path]>"
                - stop
        # check dungeon name and type
        - if !<[flag_map].get[dungeon_name].exists>:
            - narrate "dungeon name is not set"
            - stop
        - if !<[flag_map].get[dungeon_type].exists>:
            - narrate "dungeon type is not set"
            - stop
        # give the item and apply the settings as a flag
        - define slot <player.inventory.first_empty>
        - give dungeon_item slot:<[slot]>
        - inventory flag d:<player.inventory> slot:<[slot]> flag_map:<player.flag[flag_map]>
        # clear the flags and close the menu
        - flag player flag_map:<script[dungeon_data].data_key[].exclude[type]>
        - inventory close

        # discard all settings
        on player clicks dungeon_discard in dungeon_menu:
        - flag player flag_map:<script[dungeon_data].data_key[].exclude[type]>
        # update the menu
        - inventory open d:dungeon_menu


#| tasks
path_menu_task:
    type: task
    debug: false
    script:
        # get the settings that area already set
        - define flag_map <player.flag[flag_map]>
        # open the default menu
        - inventory open d:dungeon_menu
        - repeat <[flag_map].get[path_count]>:
            # for each path set the length and path item
            - inventory set slot:<[value].add[2]> d:<player.open_inventory> origin:dungeon_path
            - inventory set slot:<[value].add[11]> d:<player.open_inventory> origin:path_length
            # set the path length if it was set before
            - if <[flag_map].deep_get[path_<[value]>.path_length].if_null[0]> > 0:
                - inventory adjust slot:<[value].add[11]> d:<player.open_inventory> lore:<[flag_map].deep_get[path_<[value]>.path_length]>
        # set the other settings if they were set before
        - if <[flag_map].get[path_count].if_null[0]> > 0:
            - inventory adjust slot:1 d:<player.open_inventory> lore:<[flag_map].get[path_count]>
        - if <[flag_map].get[dungeon_name].exists>:
            - inventory adjust slot:10 d:<player.open_inventory> lore:<[flag_map].deep_get[dungeon_name]>
        - if <[flag_map].get[dungeon_type].exists>:
            - inventory adjust slot:19 d:<player.open_inventory> lore:<[flag_map].deep_get[dungeon_type]>
        - if <[flag_map].get[connect_chance].exists>:
            - inventory adjust slot:20 d:<player.open_inventory> lore:<[flag_map].deep_get[connect_chance]>

branch_menu_task:
    type: task
    debug: false
    script:
    # get the settings for the current path
    - define path_map <player.flag[flag_map.path_<player.flag[path_editing]>]>
    - inventory open d:path_menu
    # check if there are already branches set
    - if <[path_map].get[branch_amount]> > 0:
        - define branch_count <[path_map].get[branch_amount]>
        - define row 0
        # each row has 5 slots
        - while <[branch_count]> > 5:
            # populate the rows
            - repeat 5:
                # set the location and length item
                - inventory set slot:<[value].add[2].add[<[row].mul[18]>]> d:<player.open_inventory> o:branch_location
                - inventory set slot:<[value].add[11].add[<[row].mul[18]>]> d:<player.open_inventory> o:branch_length
                - define branch_index <[value].add[<[row].mul[5]>]>
                # if the branch already has settings, add them
                - if <[path_map].get[branch_<[branch_index]>].exists>:
                    - inventory adjust d:<player.open_inventory> slot:<[value].add[2].add[<[row].mul[18]>]> lore:<[path_map].deep_get[branch_<[branch_index]>.location].if_null[branch_location]>
                    - inventory adjust d:<player.open_inventory> slot:<[value].add[11].add[<[row].mul[18]>]> lore:<[path_map].deep_get[branch_<[branch_index]>.length].if_null[branch_length]>
            # go to the next row and decrease the amount of branches to create by 5
            - define row:++
            - define branch_count:-:5
        - repeat <[branch_count]>:
                - inventory set slot:<[value].add[2].add[<[row].mul[18]>]> d:<player.open_inventory> o:branch_location
                - inventory set slot:<[value].add[11].add[<[row].mul[18]>]> d:<player.open_inventory> o:branch_length
                - define branch_index <[value].add[<[row].mul[5]>]>
                - if <[path_map].get[branch_<[branch_index]>].exists>:
                    - inventory adjust d:<player.open_inventory> slot:<[value].add[2].add[<[row].mul[18]>]> lore:<[path_map].deep_get[branch_<[branch_index]>.location].if_null[branch_location]>
                    - inventory adjust d:<player.open_inventory> slot:<[value].add[11].add[<[row].mul[18]>]> lore:<[path_map].deep_get[branch_<[branch_index]>.length].if_null[branch_length]>
task_generate:
    type: task
    debug: false
    script:
    # first a template of the dungeon is generated using blocks for visual troubleshooting
    # then each block has its corrosponding room type pasted at the correct location

    #     NORTH   EAST    SOUTH   WEST
    # 4 connections crossroad
    #       x       x       x       x   COAL_BLOCK
    # 3 connections T_junction
    #       x       x       x       0   EMERALD_BLOCK
    #       x       x       0       x   DIAMOND_BLOCK
    #       x       0       x       x   GOLD_BLOCK
    #       0       x       x       x   LAPIS_BLOCK
    # 2 connections corner
    #       x       x       0       0   POLISHED_GRANITE
    #       0       x       x       0   POLISHED_ANDESITE
    #       0       0       x       x   POLISHED_DIORITE
    #       x       0       0       x   END_STONE
    # 2 connections hallway
    #       x       0       x       0   REDSTONE_BLOCK
    #       0       x       0       x   OBSIDIAN
    # 1 connection  room
    #       x       0       0       0   PRISMARINE
    #       0       x       0       0   GLASS
    #       0       0       x       0   NETHERRACK
    #       0       0       0       x   SPONGE

    # tracking current rooms for branches and path termination
    - define current_path 1
    - define current_room 1
    - define current_branch 1

    # set starting room
    - modifyblock <[dungeon_origin].add[16,40,16]> glowstone
    # generate each path
    - repeat <[path_count]>:
        # get the settings for the current path
        - define path_settings <[dungeon_settings].get[path_<[value]>]>
        #  start the new path at the starting room
        - define new_space <[dungeon_origin].add[16,40,16]>

        # check which spaces adjacent to the starting room are empty and add them to the open_list
        - if <[new_space].add[0,0,-1].material.name> == air:
            - define open_list:->:north
        - if <[new_space].add[1,0,0].material.name> == air:
            - define open_list:->:east
        - if <[new_space].add[0,0,1].material.name> == air:
            - define open_list:->:south
        - if <[new_space].add[-1,0,0].material.name> == air:
            - define open_list:->:west
        # if there is no open space no new paths can generate and the path generation stops
        - if !<[open_list].exists>:
            - repeat stop
        # choose a random direction from the list and mark that gate to be opened
        - define start_direction <[open_list].random>
        - define open_list:!
        - define open_gates:->:<[start_direction]>

        # create a main path
        - while <[current_room]> <= <[path_settings].get[path_length]>:
            #  add the previous room to the connected rooms, direction must be reversed since the start direction is relative to the previous block
            - define connected_list:->:<proc[invert].context[<[start_direction]>]>

            # set the location to the current room
            - choose <[start_direction]>:
                - case north:
                    - define new_space <[new_space].add[0,0,-1]>
                - case east:
                    - define new_space <[new_space].add[1,0,0]>
                - case south:
                    - define new_space <[new_space].add[0,0,1]>
                - case west:
                    - define new_space <[new_space].add[-1,0,0]>
            # check if the path should branch at this room
            - repeat <[path_settings].get[branch_amount]>:
                - if <[path_settings].deep_get[branch_<[value]>.location]> == <[current_room]>:
                    # add a map with the location and length to the list of branches to create for this path
                    - define branch_list:->:<map.with[location].as[<[new_space]>].with[length].as[<[path_settings].deep_get[branch_<[value]>.length]>]>
            # check if this is the last room in the path, this ensures the last room is an end
            - if <[current_room]> != <[path_settings].get[path_length]>:
                # check for empty adjacent blocks
                - inject task_open_list


                # continue if there is a place to go, else the current room will be an end
                - if <[open_list].exists>:
                    # check if there are rooms adjacent to the new room and connect them if the check for connected_chance succeeds
                    - inject task_adjacent_blocks
                    # pick a random direction for the next room
                    - define start_direction <[open_list].random>
                    # add the next room to the list of connected rooms
                    - define connected_list <[connected_list].include[<[start_direction]>]>

            # choose a room type depending on the connections
            - inject task_parse_connected
            - define connected_list:!
            # increase the room counter
            - define current_room:++
            # end the path loop if the path ended here
            - if !<[open_list].exists>:
                - while stop
            - define open_list:!

        - foreach <[branch_list].if_null[<list>]>:
            # get the branch location
            - define new_space <[value].get[location]>
            # check which spaces adjacent to the branch room are empty and add them to the open_list
            - if <[new_space].add[0,0,-1].material.name> == air:
                - define open_list:->:north
            - if <[new_space].add[1,0,0].material.name> == air:
                - define open_list:->:east
            - if <[new_space].add[0,0,1].material.name> == air:
                - define open_list:->:south
            - if <[new_space].add[-1,0,0].material.name> == air:
                - define open_list:->:west
            # if there is no space for the branch go to the next branch
            - if !<[open_list].exists>:
                - foreach next
            # pick a random starting direction
            - define start_direction <[open_list].random>
            - define open_list:!
            - define branch_room_material <[new_space].material.name>
            # check for each direction if the branch room material has that connection
            - foreach <script[connected_data].data_key[keys]>:
                - if <script[block_data].data_key[blocks_<[value]>_connected].contains[<[branch_room_material]>]>:
                    # add this direction to the connected list
                    - define connected_list:->:<[value]>
            # add the direction of the new branch to the connected_list
            - define connected_list:->:<[start_direction]>
            # choose a room type depending on the connections
            - inject task_parse_connected
            - define connected_list:!

            # track the current room being created
            - define current_room 1
            # generate the current branch
            - while <[current_room]> <= <[value].get[length]>:

                #  add the previous room to the connected rooms, direction must be reversed since the start direction is relative to the previous block
                - define connected_list:->:<proc[invert].context[<[start_direction]>]>

                # set the location to the current room
                - choose <[start_direction]>:
                    - case north:
                        - define new_space <[new_space].add[0,0,-1]>
                    - case east:
                        - define new_space <[new_space].add[1,0,0]>
                    - case south:
                        - define new_space <[new_space].add[0,0,1]>
                    - case west:
                        - define new_space <[new_space].add[-1,0,0]>
                # ensures the last room is an end
                - if <[current_room]> != <[value].get[length]>:
                    # check for empty adjacent blocks
                    - inject task_open_list

                    # continue if there is a place to go, else the current room will be an end
                    - if <[open_list].exists>:
                        # check if there are rooms adjacent to the new room and connect them if the check for connected_chance succeeds
                        - inject task_adjacent_blocks
                        # pick a random direction for the next room
                        - define start_direction <[open_list].random>
                        # add the next room to the list of connected rooms
                        - define connected_list <[connected_list].include[<[start_direction]>]>

                # choose a room type depending on connections
                - inject task_parse_connected
                - define connected_list:!
                # increase the room counter
                - define current_room:++
                # end the branch loop if the path ended here
                - if !<[open_list].exists>:
                    - while stop
                - define open_list:!
        # increase the path counter
        - define current_path:++
        # clear the branch list for the next path
        - define branch_list:!
    # generate the rooms after all the paths and branches are done
    - foreach <[room_coordinates]>:
        # rooms are generated relative to the dungeon origin at height 100
        - define room_location <[dungeon_origin].add[0,100,0]>
        # subtract the offset of the room location and multiply by the room size
        # TODO add dungeon boundaries
        # TODO change room sizes to be variable
        - define room_location:<[room_location].with_x[<[value].x.sub[16].mul[21]>]>
        - define room_location:<[room_location].with_z[<[value].z.sub[16].mul[21]>]>
        # check which material this room's block is and get the path for that room type
        - define schematic_data <proc[choose_room].context[<list[<[value].material.name>|<[dungeon_type]>]>]>
        - define schematic_path <[schematic_data].get[name]>
        - define rotation <[schematic_data].get[rotation]>

        # ensure the schematic name is unique per dungeon and per room
        - define schematic_name <[schematic_path].replace[\].with[_]>_<[rotation]>
        # load the schematic if it isnt loaded yet
        - if !<schematic[<[schematic_name]>].exists>:
            - ~schematic load name:<[schematic_name]> filename:\dungeons\<[schematic_path]>
            # rotate the schematic to face the right way
            - ~schematic rotate name:<[schematic_name]> angle:<[rotation]>
            # define the right offset for the location
        - choose <[rotation]>:
            - case 270:
                - define offset 0,0,<schematic[<[schematic_name]>].length.sub[1]>

            - case 180:
                - define offset <schematic[<[schematic_name]>].width.sub[1]>,0,<schematic[<[schematic_name]>].length.sub[1]>

            - case 90:
                - define offset <schematic[<[schematic_name]>].width.sub[1]>,0,0

            - case 0:
                - define offset 0,0,0
            - default:
                - define offset 0,0,0

        # load the area where the schematic will be pasted
        - inject load_area
        # paste the schematic at the location
        - ~schematic paste name:<[schematic_name]> <[room_location].add[<[offset]>]>
        # check if the center of the room is loaded
        # TODO add variable light locations for rooms
        - if !<[room_location].add[11,4,11].chunk.is_loaded>:
            - announce "chunloading error"
        # add a fake light to force a light update
        - light <[room_location].add[11,4,11]> 1 d:2s
        # check if the randomloot and randommob scripts are loaded
        - define randomloot_loaded <script[determine_loot].exists>
        - define randommobs_loaded <script[determine_entity].exists>
        - if <[randomloot_loaded]> || <[randommobs_loaded]>:
            # check each block in the schematic
            - note <schematic[<[schematic_name]>].cuboid[<[room_location].add[<[offset]>]>]> as:dungeon_cuboid_<[dungeon_id]>_<[room_location]>
            - foreach <schematic[<[schematic_name]>].cuboid[<[room_location].add[<[offset]>]>].blocks>:
                # ensure the block is loaded so tags dont error
                - if !<[value].chunk.is_loaded>:
                    - chunkload <[value]> duration:5s
                # check if the block has an inventory
                - if <[randomloot_loaded]> && <[value].has_inventory>:
                    # check for randomloot placeholder items and replace them with loot
                    - run determine_loot def:<[value]>
                # check if the block is a spawner
                - if <[randommobs_loaded]> && <[value].material.name> == spawner:
                    # spawn a random mob from a list and remove the spawner if it has a randommob flag
                    - flag <cuboid[dungeon_cuboid_<[dungeon_id]>_<[room_location]>]> random_spawners:->:<[value]>
        - else:
            - debug report name:randomloot "script not found, skipping loot and mob generation"
    # open the gates in the starting room
    - inject task_open_gates
    # indicate the dungeon generation script is done
    - flag <world[<[dungeon_id]>]> dungeon_spawned

task_open_gates:
    type: task
    debug: false
    script:
    # update light to fix possible lighting bugs
    - wait 1t
    - light <player.location> 1 d:1t
    # open the gates in the specified directions
    # TODO variable gate position instead of hardcoded
    - foreach <[open_gates]>:
        - choose <[value]>:
            - case north:
                - modifyblock <[dungeon_origin].add[9,102,3].to_cuboid[<[dungeon_origin].add[11,104,3]>]> air

            - case east:
                - modifyblock <[dungeon_origin].add[17,102,9].to_cuboid[<[dungeon_origin].add[17,104,11]>]> air

            - case south:
                - modifyblock <[dungeon_origin].add[9,102,17].to_cuboid[<[dungeon_origin].add[11,104,17]>]> air

            - case west:
                - modifyblock <[dungeon_origin].add[3,102,9].to_cuboid[<[dungeon_origin].add[3,104,11]>]> air

    - wait 1t
    - light <player.location> 1 d:1t
load_area:
    type: task
    debug: false
    script:
    # TODO change to allow variable room sizes
    - define min_x <[room_location].x>
    - define min_z <[room_location].z>

    - define current_x <[min_x]>
    # check how many chunks the schematic will contain
    # TODO change to <schematic[schematic].cuboid[origin].chunks>
    - if <[min_x].mod[16]> >= 7:
        - define x_iterations 3
    - else:
        - define x_iterations 2
    - if <[min_z].mod[16]> >= 7:
        - define z_iterations 3
    - else:
        - define z_iterations 2

    - repeat <[x_iterations]>:
        - define current_z <[min_z]>
        - repeat <[z_iterations]>:
            - chunkload <chunk[<[current_x].div_int[16]>,<[current_x].div_int[16]>,<[dungeon_id]>]> d:20s
            - define current_z:+:16
        - define current_x:+:16
task_open_list:
    type: task
    debug: false
    script:
    # check in each direction if there are no blocks
    - if <[new_space].add[0,0,-1].material.name> == air:
        - define open_list:->:north
    - if <[new_space].add[1,0,0].material.name> == air:
        - define open_list:->:east
    - if <[new_space].add[0,0,1].material.name> == air:
        - define open_list:->:south
    - if <[new_space].add[-1,0,0].material.name> == air:
        - define open_list:->:west

task_adjacent_blocks:
    type: task
    debug: false
    script:
    # if a direction was the starting direction it is already connected and doesnt need to be checked, if it is air there is no room to connect to
    - if <[start_direction]> != south && !<[new_space].add[0,0,-1].material.name.ends_with[air]>:
        # check if the random roll succeeds the connect chance and if its not bedrock(dungeon boundary)
        - if <util.random.int[1].to[100]> <= <[connect_chance]> && <[new_space].add[0,0,-1].material.name> != bedrock:
            # add the adjacent room to the list of connected rooms
            - define connected_list <[connected_list].include[north]>
            # change the material of the adjacent room to represent the new connection
            - choose <[new_space].add[0,0,-1].material.name>:
                - case prismarine:
                    - modifyblock <[new_space].add[0,0,-1]> redstone_block
                - case glass:
                    - modifyblock <[new_space].add[0,0,-1]> polished_andesite
                - case sponge:
                    - modifyblock <[new_space].add[0,0,-1]> polished_diorite
                - case polished_granite:
                    - modifyblock <[new_space].add[0,0,-1]> emerald_block
                - case end_stone:
                    - modifyblock <[new_space].add[0,0,-1]> gold_block
                - case obsidian:
                    - modifyblock <[new_space].add[0,0,-1]> lapis_block
                - case diamond_block:
                    - modifyblock <[new_space].add[0,0,-1]> coal_block
                # in case its the starting room, add that gate to the list
                - case glowstone:
                    - define open_gates:->:south

    # see comments above
    - if <[start_direction]> != west && !<[new_space].add[1,0,0].material.name.ends_with[air]>:
        - if <util.random.int[1].to[100]> <= <[connect_chance]> && <[new_space].add[1,0,0].material.name> != bedrock:
            - define connected_list <[connected_list].include[east]>
            - choose <[new_space].add[1,0,0].material.name>:
                - case prismarine:
                    - modifyblock <[new_space].add[1,0,0]> end_stone
                - case glass:
                    - modifyblock <[new_space].add[1,0,0]> obsidian
                - case netherrack:
                    - modifyblock <[new_space].add[1,0,0]> polished_diorite
                - case polished_granite:
                    - modifyblock <[new_space].add[1,0,0]> diamond_block
                - case polished_andesite:
                    - modifyblock <[new_space].add[1,0,0]> lapis_block
                - case redstone_block:
                    - modifyblock <[new_space].add[1,0,0]> gold_block
                - case emerald_block:
                    - modifyblock <[new_space].add[1,0,0]> coal_block
                - case glowstone:
                    - define open_gates:->:west

    # see comments above
    - if <[start_direction]> != north && !<[new_space].add[0,0,1].material.name.ends_with[air]>:
        - if <util.random.int[1].to[100]> <= <[connect_chance]> && <[new_space].add[0,0,1].material.name> != bedrock:
            - define connected_list <[connected_list].include[south]>
            - choose <[new_space].add[0,0,1].material.name>:
                - case glass:
                    - modifyblock <[new_space].add[0,0,1]> polished_granite
                - case netherrack:
                    - modifyblock <[new_space].add[0,0,1]> redstone_block
                - case sponge:
                    - modifyblock <[new_space].add[0,0,1]> end_stone
                - case polished_andesite:
                    - modifyblock <[new_space].add[0,0,1]> emerald_block
                - case polished_diorite:
                    - modifyblock <[new_space].add[0,0,1]> gold_block
                - case obsidian:
                    - modifyblock <[new_space].add[0,0,1]> diamond_block
                - case lapis_block:
                    - modifyblock <[new_space].add[0,0,1]> coal_block
                - case glowstone:
                    - define open_gates:->:north

    # see comments above
    - if <[start_direction]> != east && !<[new_space].add[-1,0,0].material.name.ends_with[air]>:
        - if <util.random.int[1].to[100]> <= <[connect_chance]> && <[new_space].add[-1,0,0].material.name> != bedrock:
            - define connected_list <[connected_list].include[west]>
            - choose <[new_space].add[-1,0,0].material.name>:
                - case prismarine:
                    - modifyblock <[new_space].add[-1,0,0]> polished_granite
                - case netherrack:
                    - modifyblock <[new_space].add[-1,0,0]> polished_andesite
                - case sponge:
                    - modifyblock <[new_space].add[-1,0,0]> obsidian
                - case polished_diorite:
                    - modifyblock <[new_space].add[-1,0,0]> lapis_block
                - case end_stone:
                    - modifyblock <[new_space].add[-1,0,0]> diamond_block
                - case redstone_block:
                    - modifyblock <[new_space].add[-1,0,0]> emerald_block
                - case gold_block:
                    - modifyblock <[new_space].add[-1,0,0]> coal_block
                - case glowstone:
                    - define open_gates:->:east

task_parse_connected:
    type: task
    debug: false
    script:
    # list all possible types of intersections
    - define block_list <script[block_data].data_key[blocks_all_connected]>
    # eliminate types based on which rooms are and arent connected
    - if <[connected_list].contains[north]>:
        # keep all blocks that have a north connection
        - define block_list <[block_list].shared_contents[<script[block_data].data_key[blocks_north_connected]>]>
    - else:
        # remove all blocks that have a north connection
        - define block_list <[block_list].exclude[<script[block_data].data_key[blocks_north_connected]>]>

    - if <[connected_list].contains[east]>:
        # keep all blocks that have a east connection
        - define block_list <[block_list].shared_contents[<script[block_data].data_key[blocks_east_connected]>]>
    - else:
        # remove all blocks that have a east connection
        - define block_list <[block_list].exclude[<script[block_data].data_key[blocks_east_connected]>]>

    - if <[connected_list].contains[south]>:
        # keep all blocks that have a south connection
        - define block_list <[block_list].shared_contents[<script[block_data].data_key[blocks_south_connected]>]>
    - else:
        # remove all blocks that have a south connection
        - define block_list <[block_list].exclude[<script[block_data].data_key[blocks_south_connected]>]>

    - if <[connected_list].contains[west]>:
        # keep all blocks that have a west connection
        - define block_list <[block_list].shared_contents[<script[block_data].data_key[blocks_west_connected]>]>
    - else:
        # remove all blocks that have a west connection
        - define block_list <[block_list].exclude[<script[block_data].data_key[blocks_west_connected]>]>

    # only 1 type should remain, otherwise something went wrong
    - if <[block_list].size> == 1:
        # set the block and add the location to the list of rooms that need to be created
        - modifyblock <[new_space]> <[block_list]>
        - define room_coordinates:->:<[new_space]>
    - else:
        - narrate "error, more than one room type"
        - while stop
#| procedures
choose_room:
    type: procedure
    debug: false
    definitions: material|dungeon_type
    script:
        # return a random room of the type corrosponding to the blocks material
        - choose <[material]>:
            - case coal_block:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\crossroad].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\crossroad\<[chosen_room]>
                - define rotation 0

            - case emerald_block:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\T_junction].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\T_junction\<[chosen_room]>
                - define rotation 270

            - case diamond_block:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\T_junction].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\T_junction\<[chosen_room]>
                - define rotation 180

            - case gold_block:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\T_junction].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\T_junction\<[chosen_room]>
                - define rotation 90

            - case lapis_block:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\T_junction].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\T_junction\<[chosen_room]>
                - define rotation 0

            - case polished_granite:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\corner].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\corner\<[chosen_room]>
                - define rotation 0

            - case polished_andesite:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\corner].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\corner\<[chosen_room]>
                - define rotation 90

            - case polished_diorite:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\corner].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\corner\<[chosen_room]>
                - define rotation 180

            - case end_stone:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\corner].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\corner\<[chosen_room]>
                - define rotation 270

            - case obsidian:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\hallway].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\hallway\<[chosen_room]>
                - define rotation 90

            - case redstone_block:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\hallway].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\hallway\<[chosen_room]>
                - define rotation 0

            - case prismarine:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\room].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\room\<[chosen_room]>
                - define rotation 0

            - case glass:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\room].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\room\<[chosen_room]>
                - define rotation 90

            - case netherrack:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\room].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\room\<[chosen_room]>
                - define rotation 180

            - case sponge:
                - define chosen_room <server.list_files[\schematics\dungeons\<[dungeon_type]>\room].random.before[.schem]>
                - define schematic_path <[dungeon_type]>\room\<[chosen_room]>
                - define rotation 270

        - determine <map.with[name].as[<[schematic_path]>].with[rotation].as[<[rotation]>]>

invert:
    type: procedure
    definitions: direction
    debug: false
    script:
    # determine the oposite direction
    - choose <[direction]>:
        - case north:
            - determine south
        - case east:
            - determine west
        - case south:
            - determine north
        - case west:
            - determine east

valid_type_proc:
    type: procedure
    definitions: type
    script:
    # validate the dungeon type preset
    # get a list of all room types
    - define type_list <script[dungeon_type_data].data_key[room_shapes]>
    - foreach <[type_list]>:
        # check if there is a .schem file in the folder for this room type
        - if <server.list_files[schematics\dungeons\<[type]>\<[value]>].filter_tag[<[filter_value].ends_with[.schem]>].if_null[<list>].is_empty>:
            - narrate "invalid, dungeon type is missing <[value]> schematic"
            - determine error
    - determine success

#| menus
dungeon_menu:
    type: inventory
    gui: true
    inventory: chest
    size: 27
    slots:
    - "[path_count] [] [] [] [] [] [] [] [dungeon_confirm]"
    - "[dungeon_name] [] [] [] [] [] [] [] []"
    - "[dungeon_type] [connect_chance] [] [] [] [] [] [] [dungeon_discard]"

path_menu:
    type: inventory
    gui: true
    inventory: chest
    size: 54
    slots:
    - "[branch_amount] [] [] [] [] [] [] [] [dungeon_confirm]"
    - "[] [] [] [] [] [] [] [] []"
    - "[] [] [] [] [] [] [] [] []"
#| items
path_count:
    type: item
    material: black_stained_glass_pane
    display name: <white>Path count
    lore:
        - "amount of paths from the starting room"
dungeon_path:
    type: item
    material: black_stained_glass_pane
    display name: <white>Dungeon path
    lore:
        - "click to configure"
path_length:
    type: item
    material: gray_stained_glass_pane
    display name: <white>Path length
    lore:
        - "length of this path"
branch_amount:
    type: item
    material: yellow_stained_glass_pane
    display name: <white>Amount of branches
    lore:
        - "ammount of branches"
branch_location:
    type: item
    material: orange_stained_glass_pane
    display name: <white>Branch location
    lore:
        - "amount of paths from the starting room"
branch_length:
    type: item
    material: red_stained_glass_pane
    display name: <white>branch length
    lore:
        - "amount of paths from the starting room"
connect_chance:
    type: item
    material: blue_stained_glass_pane
    display name: <white>Connection chance
    lore:
        - "Chance for unconnected adjacent rooms to connect"
dungeon_name:
    type: item
    material: brown_stained_glass_pane
    display name: <white>Dungeon name
    lore:
        - "Dungeon display name"
dungeon_type:
    type: item
    material: purple_stained_glass_pane
    display name: <white>Dungeon room type
    lore:
        - "Dungeon room type, available types:"
        - "<server.list_files[schematics/dungeons].comma_separated.if_null[ERROR_NO_DUNGEONS_FOUND]>"

dungeon_confirm:
    type: item
    material: lime_stained_glass_pane
    display name: <white>Confirm
    lore:
        - "create the dungeon with the set parameters"
dungeon_discard:
    type: item
    material: red_stained_glass_pane
    display name: <white>Discard settings
    lore:
        - "Discard all settings and start over"


dungeon_item:
    type: item
    material: oak_sign
    display name: <white>Dungeon placer
    lore:
        - "right click a block to make it a dungeon portal"

#| entities
dungeon_marker:
    type: entity
    entity_type: armor_stand
    has_ai: false
    marker: true
    visible: false
    gravity: false
    custom_name_visible: true


#| commands
dungeon:
    type: command
    name: dungeon
    description: dungeon setup
    usage: /dungeon
    permission: imperial.dungeon.create
    permission message: You dont have permission to use this command
    script:
        - if !<player.has_flag[flag_map]>:
            - flag player flag_map:<script[dungeon_data].data_key[].exclude[type]>
        - inject path_menu_task
#| data scripts
dungeon_data:
    type: data
    path_1:
        branch_amount: 0
        path_length: 0
    path_2:
        branch_amount: 0
        path_length: 0
    path_3:
        branch_amount: 0
        path_length: 0
    path_4:
        branch_amount: 0
        path_length: 0
    connect_chance: 0
    path_count: 0
dungeon_type_data:
    type: data
    room_shapes:
        - corner
        - hallway
        - room
        - T_junction
        - start
        - crossroad
connected_data:
    type: data
    keys:
        - north
        - east
        - south
        - west
block_data:
    type: data
    blocks_all_connected:
        - polished_granite
        - polished_diorite
        - polished_andesite
        - end_stone
        - lapis_block
        - diamond_block
        - emerald_block
        - gold_block
        - coal_block
        - obsidian
        - netherrack
        - glass
        - sponge
        - prismarine
        - redstone_block

    blocks_north_connected:
        - prismarine
        - polished_granite
        - end_stone
        - redstone_block
        - gold_block
        - diamond_block
        - emerald_block
        - coal_block

    blocks_east_connected:
         - glass
         - polished_granite
         - polished_andesite
         - obsidian
         - lapis_block
         - diamond_block
         - emerald_block
         - coal_block

    blocks_south_connected:
        - netherrack
        - polished_andesite
        - polished_diorite
        - redstone_block
        - lapis_block
        - gold_block
        - emerald_block
        - coal_block

    blocks_west_connected:
        - sponge
        - polished_diorite
        - end_stone
        - obsidian
        - lapis_block
        - gold_block
        - diamond_block
        - coal_block



